/*
 * FILE --- llp.cc
 *
 * PURPOSE --- Source file for a class that implements a Limited Lookahead Policy (LLP)
 * controller. The user specifies the dimension of the state-space, the dimension of the
 * control space, the lookahead horizon (i.e., the number of steps to lookahead), and
 * the total number of possible control inputs. The number of possible control inputs
 * needs to be finite so that the optimization is duable. Morover, the user has to 
 * provide the dynamics of the underlying control system, the cost function based on
 * which the optimization is performed, and all of the possible control inputs so that
 * proper optimization is performed. Given the current system state the algorithm 
 * generates all the possible future states within the lookahed horizon. Then it returns 
 * the control value of the first control input of the path with the smalles cost.
 *
 */

// Include Files
#include "llp.hh"

// Define this macro if you want to use this utility without linking
// to the RCS Library
#ifdef NO_RCS
#include <stdio.h>            /* printf function */
#else
#include "rcs_prnt.hh"          /* rcs_print_ functions */
#endif

/*************************************************************/

LLP::LLP(int _plant_order, int _ctrl_order, 
	 int _lookahead_horizon, int _num_diff_ctrl)
{
  // No error at start
  llp_err = LLP_NO_ERROR;
  
  // Initialize plant parameters
  plant_order = _plant_order;
  ctrl_order = _ctrl_order;
  lookahead_horizon = _lookahead_horizon;  
  num_diff_ctrl= _num_diff_ctrl;
  
  // Allocate memory 
  allocate_memory();
  
  // Check and report if error occured
  if (llp_err != LLP_NO_ERROR)
    {
#ifdef NO_RCS
      printf("LLP: ERROR --- Not Enought Memory!!!\n");
      printf("The LLP controller will not function correctly ...\n");
#else
      rcs_print("LLP: ERROR --- Not Enought Memory!!!\n");
      rcs_print("The LLP controller will not function correctly ...\n");
#endif
    }

  // Set all the control inputs to zero
  set_ctrl_inputs();
}

LLP::~LLP()
{
  // Free the allocated memory
  free_memory();
}


void LLP::allocate_memory(void)
{
  if ((plant_state = new double[plant_order]) == NULL)
    llp_err = LLP_MEM_ERROR;
  
  if ((trial_state = new double[plant_order]) == NULL)
    llp_err = LLP_MEM_ERROR;
  
  if ((next_state = new double[plant_order]) == NULL)
    llp_err = LLP_MEM_ERROR;
  
  if ((ctrl_inputs = new double*[num_diff_ctrl]) == NULL)
    llp_err = LLP_MEM_ERROR;
  
  for (int ii = 0; ii < num_diff_ctrl; ii++)
    {
      if ((ctrl_inputs[ii] = new double[ctrl_order]) == NULL)
	llp_err = LLP_MEM_ERROR;
    }
}


void LLP::free_memory(void)
{
  delete [] plant_state;
  delete [] trial_state;
  delete [] next_state;
  
  for (int ii = 0; ii < num_diff_ctrl; ii++)
    {
      delete [] ctrl_inputs[ii];
    }  
  delete [] ctrl_inputs;
}


void LLP::find_ctrl(double *_ctrl)
{
  int ii, jj, kk;   // Index variables for the loops
  double J = 0, J_min = 0; // Conveniance variables for the cost
  
  // Create new lists to hold the current and next (stage) nodes
  current_list = new open_list();
  next_list = new open_list();

  // Make the current plant state the only node in the current list
  temp_node = new open_node(plant_order);
  temp_node->set_node(plant_state, LLP_INVD_CTRL, 0);
  current_list->add_node(temp_node);

  // Expend all possible states within the lookahead horizon
  for(ii = 0; ii < lookahead_horizon; ii++)
    {
      // Start with the first node of the curent list (stage)
      current_node = current_list->firstPtr;

      // Visit all the elements in the list
      for(jj = 0; jj < current_list->num_of_elements; jj++)
	{
	  // Get the state of the current node
	  current_node->get_state(trial_state);

	  // Given the state and the possible control inputs
	  // find all possible next states
	  for(kk = 0; kk < num_diff_ctrl; kk++)
	    {
	      // Calculate the next state
	      dynamics(trial_state, ctrl_inputs[kk], next_state);
	      
	      // Calculate the cost of moving to the next state
	      J = current_node->get_cost() + find_cost(next_state, ctrl_inputs[kk]);
	     
	      // Add the new next state to the list of next states
	      temp_node = new open_node(plant_order);
	      if (ii == 0)
		{
		  // If first stage then save the current control
		  temp_node->set_node(next_state, kk, J);
		}
	      else
		{
		  // If higher stage save the control of the parent
		  temp_node->set_node(next_state, current_node->get_ctrl(), J);		  
		}
	      next_list->add_node(temp_node);
	    }
	  
	  // Move to the next node
	  current_node = current_node->nextPtr;
	}

      // Move to the next stage
      temp_list = current_list;
      current_list = next_list;
      next_list = new open_list();

      // Free the unnecessary nodes
      delete temp_list;
    }

  // Find the node with the least cost
  // Start with the first node
  current_node = current_list->firstPtr;
  J_min = current_node->get_cost();
  temp_node = current_node;
  current_node = current_node->nextPtr;

  // Search through all the nodes in the list
  for(jj = 1; jj < current_list->num_of_elements; jj++)
    {
      if (current_node->get_cost() < J_min)
	{
	  // Save the cost and the node
	  J_min = current_node->get_cost();
	  temp_node = current_node;
	}

      // Move to the next node
      current_node = current_node->nextPtr;
    }

  // Given the index of the best control retrieve its value
  for(ii = 0; ii < ctrl_order; ii++)
    {
      _ctrl[ii] = ctrl_inputs[temp_node->get_ctrl()][ii];
    }

  // Free the allocated memory within this function
  delete current_list;
  delete next_list;
}

void LLP::set_ctrl_inputs(const double **_ctrl_inputs = NULL)
{
  int ii, jj;   // Index variable

  // If no value specified set all the control inputs to zero
  if (_ctrl_inputs == NULL)
    {
      for(ii = 0; ii < num_diff_ctrl; ii++)
	{
	  for(jj = 0; jj < ctrl_order; jj++)
	    {
	      ctrl_inputs[ii][jj] = 0.0;
	    }
	}
    }
  else // Set the control inputs to the specified value
    {
      for(ii = 0; ii < num_diff_ctrl; ii++)
	{
	  for(jj = 0; jj < ctrl_order; jj++)
	    {
	      ctrl_inputs[ii][jj] = _ctrl_inputs[ii][jj];
	    }
	}
    }
}

void LLP::set_ctrl_input(int _input_num, const double *_ctrl_input = NULL)
{
  // If no value specified set the given control input to zero
  if (_ctrl_input == NULL)
    {
      for(int jj = 0; jj < ctrl_order; jj++)
	{
	  ctrl_inputs[_input_num][jj] = 0.0;
	}
    }
  else // Set the given control input to the specified value
    {
      for(int jj = 0; jj < ctrl_order; jj++)
	{
	  ctrl_inputs[_input_num][jj] = _ctrl_input[jj];
	}
    }
}

